/*   -*- C++ -*-
 *   Island, converted from the Acorn BBC computer version (BASIC) to C & GGI
 *   by W.H.Scholten (wouters@cistron.nl), february 1999.
 *
 *   The BBC BASIC version appeared in the book:
 *      Creative graphics on the BBC microcomputer, John Cownie, 1982
 *
 *   This software is placed in the public domain and can be used freely
 *   for any purpose. It comes without any kind of warranty, either
 *   expressed or implied, including, but not limited to the implied
 *   warranties of merchantability or fitness for a particular purpose.
 *   Use it at your own risk. the author is not responsible for any damage
 *   or consequences raised by use or inability to use this program.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ggi/ggi.h>
#include <math.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>

ggi_visual_t vis = NULL;
const ggi_directbuffer *dbuf; // not used

int  origin_x = 0;
int  origin_y = 0;

#ifndef False
#  define False 0
#endif
#ifndef True
#  define True 1
#endif

int	stride;	        // virtual screen width, in bytes
int	width;	        // visible screen width, in pixels
int	height;	        // visible screen height, in pixels

void fail(char *reason) { 
	fprintf (stderr, "%s", reason); 
	if (vis != NULL) {
		ggiClose (vis);
		ggiExit ();
	}
	exit(1); 
}


void init_graphics(void) {
  	ggi_mode m;
	ggiInit();

	if (!(vis = ggiOpen (NULL)) )
		fail ("Error opening GGI display.\n");

	ggiSetFlags(vis, GGIFLAG_ASYNC);

	if (ggiSetSimpleMode(vis, GGI_AUTO, GGI_AUTO, GGI_AUTO, GT_8BIT)) {
		fprintf (stderr, "Error switching to %dx%d 8 bpp mode.\n", width, height);
		fail ("");
	}

	ggiGetMode(vis, &m);

	width = m.virt.x;
	height = m.virt.y;

	if (!(dbuf = ggiDBGetBuffer (vis, 0)) )
		fail ("Error getting display buffer.\n");

	if (dbuf->layout != blPixelLinearBuffer)
		fail ("Error: nonlinear display buffer\n");

	if (dbuf->layout != blPixelLinearBuffer)
		fail ("Error: nonlinear display buffer\n");

	if (! (dbuf->type & GGI_DB_SIMPLE_PLB))
		fail ("Error: non-standard display buffer\n");

	stride = dbuf->buffer.plb.stride;
}


void close_graphics(void) {
	ggiClose (vis);
	ggiExit ();
}


// rgba
ggi_color mypalette[16] = {
	     0,      0,       0, 255<<8, /* 0 = black */
	255<<8,      0,       0, 255<<8, /* 1 = red */
	     0, 255<<8,       0, 255<<8, /* 2 = green */
	255<<8, 255<<8,       0, 255<<8, /* 3 = yellow */
	     0,      0,  255<<8, 255<<8, /* 4 = blue */
	255<<8,      0,  255<<8, 255<<8, /* 5 = magenta */
         0, 255<<8,  255<<8, 255<<8, /* 6 = cyan */
    255<<8, 255<<8,  255<<8, 255<<8, /* 7 = white */
};


void draw_line(int x1, int y1, int x2, int y2) {
	x1 = (x1+origin_x)/4;
	y1 = 256 - (y1+origin_y)/4;
	x2 = (x2+origin_x)/4;
	y2 = 256 - (y2+origin_y)/4;

	ggiDrawLine(vis, x1, y1, x2, y2);
}


void draw_triangle(int x1, int y1, int x2, int y2, int x3, int y3) {
	int  j;

	x1 = (x1+origin_x)/4;
	y1 = 256 - (y1+origin_y)/4;
	x2 = (x2+origin_x)/4;
	y2 = 256 - (y2+origin_y)/4;
	x3 = (x3+origin_x)/4;
	y3 = 256 - (y3+origin_y)/4;

	// bad quick hack
#define TRIANGDIV 50
	for (j=0; j <= TRIANGDIV; j++) {
		ggiDrawLine(vis, x1, y1, x2+((x3-x2)*j)/TRIANGDIV, y2+((y3-y2)*j)/TRIANGDIV);
	}
}


void draw_sea(int yl) {
	int  prev2_x, prev2_y, prev_x, prev_y, new_x, new_y;
	int  i, x, y, ny;
	int  pixel;
	int  s, cc;

	float sin_table[101];

	ggiSetPalette(vis, 0, 8, &mypalette[0]);

	for (i=8; i<=15; i++) {
		ggi_color mycolour;

		// set colours 8-15 to blue
		mycolour.r = 0;
		mycolour.g = 0;
		mycolour.b = 255<<8;
		ggiSetPalette(vis, i, 1, &mycolour);
	}
	for (i=0; i<=100; i++) {
		sin_table[i] = sin((float) i);
	}

	for (y=0; y <= yl; y += 12) {
		origin_x = 0;
		origin_y = y;

		s = 8 +(rand() %10) +1;
		prev_x = 0;
		prev_y = s * sin_table[y % 100];
		cc = 7 +(rand() % 8) +1;
		for (x=0; x<=1279; x+=24) {
			pixel = cc;
			cc++; if (cc==16) cc=8;

			ggiSetGCForeground(vis, pixel);
			new_x = x;
			new_y = s*sin_table[(x+y) % 100];

			draw_line(prev_x, prev_y, new_x, new_y);

			prev_x = new_x;
			prev_y = new_y;
		}
	}

	yl = y;
	s = 8;

	origin_x = 0;
	origin_y = yl;

	new_x = 0;
	new_y = s*sin_table[yl %100];
	prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

	for (x=0; x <= 1279; x += 24) {
		ny = s*sin_table[(x+24+yl) % 100];

		pixel = 6;
		ggiSetGCForeground(vis, pixel);

		y = s*sin_table[(x+yl) % 100];

		new_x = x;
		new_y = 1200;

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		new_x = x+24;
		new_y = 1200;

		draw_triangle(prev2_x, prev2_y, prev_x, prev_y, new_x, new_y);

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		new_x = x+24;
		new_y = ny;
		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		new_x = x;
		new_y = y;

		draw_triangle(prev2_x, prev2_y, prev_x, prev_y, new_x, new_y);

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		pixel = cc;
		cc++; if (cc==16) cc=8;

		ggiSetGCForeground(vis, pixel);

		new_x = x+24;
		new_y = ny;

		draw_line(prev_x, prev_y, new_x, new_y);

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;
	}
}


void draw_isle(int xo, int yo, int s) {
	float a, b;
	int  prev2_x, prev2_y, prev_x, prev_y, new_x, new_y;
	int  x, y;
	int  s1;

	ggiSetGCForeground(vis, 3); // yellow
	origin_x = xo;
	origin_y = yo;

	prev_x = 0; prev_y = 0;
	new_x =0; new_y = 0;

	for (a =0; a<2*M_PI+0.2; a+=.2) {
		x = s*sin(a); y = s*cos(a);
		if (y>0) y /= 2; else y /= 3;

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;
		new_x=0; new_y=0;
		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		new_x = x;
		new_y = y;

		draw_triangle(prev2_x, prev2_y, prev_x, prev_y, new_x, new_y);

	}

	ggiSetGCForeground(vis, 2); // green
	for (a =M_PI/2; a<M_PI; a+=.05) {
		s1= s *sin(a);
		origin_x = xo;
		origin_y = yo+(cos(a)*s)/3;
		new_x = -s1; new_y = 0;
		prev_x  = new_x;  prev_y  = new_y;
		for (b= -M_PI/2; b<M_PI/2+.2; b +=.2) {
			new_x = s1*sin(b);
			new_y = s1*cos(b)/2;

			draw_line(prev_x, prev_y, new_x, new_y);

			prev_x  = new_x;  prev_y  = new_y;
		}
	}
}


void draw_palm (int x, int y, int s) {
	float st, a, si, co;
	int  prev2_x, prev2_y, prev_x, prev_y, new_x, new_y;
	int  i, s1;

	origin_x = x - s;
	origin_y = y;

	new_x = s;
	new_y = 0;
	prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

	s1 = 1.1 *s;
	st = 40.0 / s;

	// draw the trunk of the palm tree
	for (a=0; a < M_PI/3; a += st) {
		ggiSetGCForeground(vis, 1);
		si = sin(a); co = cos(a);
		prev2_x = s*co; prev2_y = s*si;
		prev_x = s1*co; prev_y = s1*si;

		new_x = s1*cos(a+2*st);
		new_y = s1*sin(a+2*st);

		draw_triangle(prev2_x, prev2_y, prev_x, prev_y, new_x, new_y);

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		new_x = s*cos(a+st);
		new_y = s*sin(a+st);
		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		new_x = s*co;
		new_y = s*si;

		draw_triangle(prev2_x, prev2_y, prev_x, prev_y, new_x, new_y);

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

		ggiSetGCForeground(vis, 7);

		new_x = s1*co;
		new_y = s1*si;

		draw_line(prev_x, prev_y, new_x, new_y);

		prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;
	}

	// now draw the leaves
	origin_x = x+s*(cos(M_PI/3)-1);
	origin_y = y+s*(sin(M_PI/3)) + 0.05*s;
	ggiSetGCForeground(vis, 2);

	new_x = 0;
	new_y = 0;
	prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;
	new_x = 0;
	new_y = 4;
	prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;

	for (a=0; a<2*M_PI; a+=.8) {
		for (i=0; i < s; i += 16) {
			if (rand() & 1)
				ggiSetGCForeground(vis, 0); // black
			else
				ggiSetGCForeground(vis, 2); // green

			new_x = i/2 *sin(a) + (i>=4 ? (rand() % (i/4) +1) : 0) - i/8;
			new_y = i/3 *cos(a) + (i>=4 ? (rand() % (i/4) +1) : 0) - i/8;
			draw_triangle(prev2_x, prev2_y, prev_x, prev_y, new_x, new_y);

			prev2_x = prev_x; prev2_y = prev_y; prev_x = new_x; prev_y = new_y;
		}
	}
}


//#define DELAY 120000 /* original delay .12s */
#define DELAY 100000
void surf() {
	int  c;
	int  i, j, k;
	int  r;
  	struct timeval tv = { 0, 0 };

	c = 8;
	for (i=0; i<=100; i++) {
		ggi_color mycolour;

		// white
		mycolour.r = 255<<8;
		mycolour.g = 255<<8;
		mycolour.b = 255<<8;
		ggiSetPalette(vis, c, 1, &mycolour);
		ggiFlush(vis);
		if (ggiEventPoll(vis, emKeyPress, &tv)) return;
		usleep(DELAY);

		// blue
		mycolour.r = 0;
		mycolour.g = 0;
		mycolour.b = 255<<8;
		ggiSetPalette(vis, c, 1, &mycolour);
		c++; if (c==16) c=8;
	}

	for (j=0; j<=100; j++) {
		r = (rand() % 5) + 1;
		for (k=0; k<=20; k++) {
			c++; if (c==16) c=8;
			ggiSetPalette(vis, c, 1, &mypalette[4]);
			ggiSetPalette(vis, ((c+r) & 7) +8, 1, &mypalette[7]);
			ggiFlush(vis);
			if (ggiEventPoll(vis, emKeyPress, &tv)) return;
			usleep(DELAY);
		}
	}
}


int main (int argc, char **argv) {
	init_graphics();

	draw_sea(650);
	draw_isle(700, 400, 250);
	draw_palm(700, 500, 300);
	draw_palm(800, 450, 300);

	surf();

	close_graphics();

	return 0;
}
